import openpyxl
import os
from base64 import b64encode
import plotly.graph_objects as go

def create_pie_analytics(df, title, column, value_column, style="height: 230px;width: 300px;"):
    layout = go.Layout(
        autosize=True,
        showlegend=False,
        title_text="<b>" + title + "</b>",
        title_x=0.5,
        title_y=0.98,
    )
    fig = go.Figure(data=[go.Pie(labels=df[column].tolist(),
                    values=df[value_column].values.tolist(), pull=[0.1, 0.05, 0.03, 0],
                    hole=.3, texttemplate="%{label}: %{value} <br>(%{percent})", insidetextorientation='horizontal', sort=True)], layout=layout)
    fig.update_traces(marker=dict(line=dict(color='white', width=3)))
    fig.update_layout(font=dict(family="sans-serif",
                      size=16, color="black"))
    img = fig.to_image(format="png", engine="kaleido")

    summary = '<img alt="execution summary" id="box" style="'+style+'" src="data:image/png;base64, {}" >'.format(
        b64encode(img).decode("utf-8"))
    return summary + df.to_html(
        table_id="table",
        index=False,
        render_links=False,
        escape=True,
    )


def getStartEndFromExcel(dirname, filename, key):
    dataframe = openpyxl.load_workbook(os.path.join(
        dirname, filename))

    dataframe1 = dataframe.active
    for row in range(0, dataframe1.max_row):
        for col in dataframe1.iter_cols(1, 1):
            if col[row].value == key:
                start = row
                break
        else:
            continue
        break
    for row in range(start, dataframe1.max_row):
        for col in dataframe1.iter_cols(1, 1):
            if col[row].value is None:
                end = row
                break
        else:
            continue
        break
    return start, end, dataframe1.max_row



import pandas
import requests
import json
import os
import sys
from functions import *
from commands_api import *
osVersions = []
from integrations.google_analytics.g_analytics import *

mobile_filename = 'web-portal-30-days-mobile-os.csv'
web_filename = "web-portal-last-12-months-tech copy.xlsx"
client_logo = "https://www.medanswering.com/images/MAS_web.png"
devices = get_devices_list()
prepare_google_analytics_report(mobile_filename, web_filename, client_logo, devices)



# osVersions = get_os_versions_list(devices)
# osVersions_major = get_os_version_major(osVersions)
# print(osVersions_major)

# print(find_nearest_decimal(13.7, osVersions_major))



from .g_functions import *
from .g_html import prepare_report
import pandas as pd
import os
import sys
sys.path.append(os.path.dirname(os.path.realpath(__file__)))

# variables:

mobile_os = ['iOS', 'Android']
web_os = ['Windows', 'Macintosh']
mobile_browsers = ['Safari', 'Chrome']
users_old = 'Active users'
os_version_old = 'Operating system with version'
users = 'Users'
os_version = 'OS Version'
browser_version = 'Browser Version'
browser = 'Browser'
version = 'Browser version'
top_count = 5
recommend = 5
web_recommend = 4
#mobile
mobile_list = [os_version, browser,  browser_version]
mobile_web = []
mobile_web_recommender = []
# web
web = []
web_recommender = []
keys = ["Operating system", "Browser", "Screen resolution"]

def prepare_google_analytics_report(mobile_filename, web_filename, client_logo, devices):
    full_path = os.path.realpath(__file__)
    dirname = os.path.dirname(
        full_path)
    # remove existing html files
    for root, dirnames, filenames in os.walk(dirname):
        for filename in filenames:
            if filename.endswith('.html'):
                os.remove(os.path.join(root, filename))
    report_filename = os.path.join(dirname, 'Google_Analytics.html')

    # mobile web
    df = pd.read_csv(os.path.join(os.getcwd(), mobile_filename),
                    skiprows=6, skip_blank_lines=True, low_memory=False, index_col=False)
    df = df.iloc[1:, :]
    df = df.rename(columns={os_version_old: os_version, users_old: users})
    df[browser_version] = df[browser] + ' ' + df[version]

    for platform in mobile_os:
        for idx, key in enumerate(mobile_list):
            new_df = df[df[os_version].str.startswith(platform, na=False)].groupby([key])[users].sum().reset_index(
            ).sort_values(by=[users], ascending=False).head(top_count)  # top count
            if new_df.columns[0] == os_version:  # remove platform name from OS version
                new_df[os_version] = new_df[os_version].str.replace(
                    platform, '').str.strip()
            new_df["%"] = round((new_df[users] / new_df[users].sum())* 100, 0).astype(int).astype(str) + '%'
            mobile_web.append(create_pie_analytics(new_df, key, key, users))
            new_df = new_df.drop(columns=[users]).reset_index(
                drop=True)  # removing users column
            # need to combine two columns of different df
            if new_df.columns[0] in [os_version, browser]:
                if idx == 0:
                    old_df = new_df.head(recommend)
                else:
                    new_df[~new_df[browser].isin(mobile_browsers)] = "" # remove unsupported browsers
                    while len(new_df) < recommend: #create empty rows to match recommendations
                        new_df.loc[len(new_df)] = ''
                    merged_df = old_df.merge(new_df.head(
                        recommend), left_index=True, right_index=True)
                    merged_df = merged_df.drop(columns=['%_x', '%_y']).reset_index(drop=True)
                    mobile_web_recommender.append(merged_df.to_html(table_id="table",
                                                                    index=False,
                                                                    render_links=False,
                                                                    escape=False,
                                                                    ))



    for idx, key in enumerate(keys):
        start, end, total_rows = getStartEndFromExcel(dirname, web_filename, key)
        df = pd.read_excel(os.path.join(dirname, web_filename), skiprows=start,
                        skipfooter=(total_rows-end), index_col=False)
        new_df = df.groupby(df.columns[0])[df.columns[1]].sum().reset_index(
        ).sort_values(by=[df.columns[1]], ascending=False).head(top_count)
        new_df["%"] = round((new_df[users] / new_df[users].sum())* 100, 0).astype(int).astype(str) + '%'
        web.append(create_pie_analytics(new_df, key, key, df.columns[1]))
        new_df = new_df.drop(columns=[users]).reset_index(
            drop=True)  # removing users column
        if idx == 0:
            one = new_df.head(web_recommend)
        if idx == 1:
            two = new_df.head(web_recommend)
        if idx == 2:
            merged_df = one.merge(new_df.head(
                web_recommend), left_index=True, right_index=True)
            merged_df = two.merge(merged_df.head(
                web_recommend), left_index=True, right_index=True)
            merged_df = merged_df[keys]
            # merged_df = merged_df[merged_df[keys[0]].isin(web_os)] # picks just Web OS
            web_recommender.append(merged_df.to_html(table_id="table",
                                                    index=False,
                                                    render_links=False,
                                                    escape=False,
                                                    ))

    # prepare report
    html = prepare_report(client_logo, mobile_web, web,
                        str(top_count), mobile_web_recommender, web_recommender, devices)
    with open(report_filename, 'w') as f:
        f.write(html)



import sweetviz as sv
import os
from bs4 import BeautifulSoup


def update(client_logo, report_filename, file):
    with open(file) as inf:
        txt = inf.read()
        soup = BeautifulSoup(txt, "html.parser")
        soup.prettify(formatter="minimal")

    # remove summary
    elements = soup.findAll(attrs={'id': ['summary-df']})
    try:
        for element in elements:
            element.decompose()
    except:
        pass

    # remove top section
    for tag in soup.findAll(attrs={'class': ['page-column-main']}):
        tag['style'] = "top: -160px;z-index:-1;"

    for tag in soup.findAll(attrs={'class': ['graph-legend', 'pos-logo-group']}):
        tag.decompose()
    
    heading = """
    <style>
        body {
            height: 0px;
        }
        
        #images {
            background: linear-gradient(to right, #ffb35a70, #E3F3FE, #E3F3FE, #aacfe8 90.33%, #ffb35a70);
            box-shadow: 0 1px 6px rgba(0, 0, 0, 0.12), 0 1px 4px rgba(0, 0, 0, 0.24);
            height: 35px;
            margin: auto;
            display: flex;
            justify-content: center;
            padding: 2px 1px 2px 10px;
            }
    </style>
    <div id="images">
        <img src="https://www.perfecto.io/sites/default/themes/custom/perfecto/logo.svg?height=57&width=200" alt="Perfecto" id="perfecto" class="center">
        <img style="margin-left: 20px;" src='"""+client_logo+"""' alt="logo" id="clientlogo" class="center">
        <a href='./""" + report_filename + """' target="_blank" ><img style="margin-left: 20px;" width="40" height="35" src="https://img.icons8.com/fluency/48/graph-report.png" id="reporticon" alt="analytics-report" /></a>
    </div> """
    with open(file, "w") as outf:
        outf.write(heading)
        outf.write(str(soup))


def generate_advanced_analytics(df, client_logo, report_filename, configpath, filename, internal):
    profile_df = df.copy(deep=False)
    del_cols = [
        col for col in df.columns if 'artifacts' in col or 'selectionCriteriaV2' in col or 'customFields' in col or 'tags' in col or 'videos' in col or 'executionEngine' in col or 'parameters' in col]
    del_cols.extend(["id", "index", "externalId", "df_index", "Duration", 'owner', 'uxDuration', 'month', 'week', 'startDate', "startTime",
                    "endTime", "job/number", "reportURL", "platforms/0/deviceId", 'platforms/0/mobileInfo/imei', 'platforms/0/mobileInfo/distributor', 'platforms/0/mobileInfo/firmware', 'triggerType', 'retry', 'hidden', 'duration', 'uiOptions/strStatus', 'version', 'testExecutionId', 'tracking/createdAt', 'tracking/lastUpdatedAt', 'failureReason/id', 'failureReason/isSystemAssigned', 'errorAnalysis/normalizedCleanException'])
    profile_df.drop(del_cols, axis=1, errors='ignore', inplace=True)
    profile_df.reset_index(drop=True, inplace=True)
    if internal == "true":
        profile_df.rename(columns={'failureReason/name': 'Custom Failure Reasons'})
    else:
         profile_df.rename(columns={'failureReasonName': 'Custom Failure Reasons'})
    profile_df.rename(columns={'name': 'Test Name', 'owner': 'Test Owner', 'message': 'Test Failure Message',
                               'automationFramework': 'Test Automation Framework', 'status': 'Status', 'platforms/0/screenResolution': 'Platform Screen Resolution', 'platforms/0/browserInfo/browserType': 'Browser Type', 'platforms/0/browserInfo/browserVersion': 'Browser Version', 'job/name': 'Job Name', 'job/branch': 'Job Branch', 'platforms/0/osVersion': 'Platform OS Version', 'platforms/0/os': 'Platform OS', 'platforms/0/mobileInfo/phoneNumber': 'Device Phone Number',
                               'model': 'Device Model', "cleanException": 'Failed Steps', 'browserType': 'Browser Type', 'browserVersion': 'Browser Version', 'platforms/0/location': 'Platform Location', 'platforms/0/deviceType': 'Platform DeviceType', 'platforms/0/mobileInfo/manufacturer': 'Device Manufacturer', 'project/version': 'Project Version', 'project/name': 'Project Name', 'platforms/0/mobileInfo/operator': 'Mobile Operator', 'platforms/0/mobileInfo/hasAudio': 'Mobile Device Audio?'}, inplace=True)
    output_file = os.path.join(os.getcwd(),  filename)
    feature_config = sv.FeatureConfig(skip=[])
    my_report = sv.analyze(profile_df, pairwise_analysis='off',
                           feat_cfg=feature_config)
    sv.config_parser.read(os.path.join(configpath, "configs", "Override.ini"))
    my_report.show_html(filepath=output_file,
                        open_browser=False,
                        layout='widescreen',
                        scale=None)
    update(client_logo,report_filename, output_file)
    print("Analytics: file://" + output_file)



import asyncio
from multiprocessing import freeze_support, Process
import threading
import glob
from argparse import ArgumentParser
import datetime
import ssl
import numpy as np
import json
import pandas
import re
import os
import sys
sys.path.append(os.path.dirname(os.path.realpath(__file__)))

pandas.options.mode.copy_on_write = True

# Variables
reportTag = ""


def customReporter(client_logo, tags_to_remove, automation_owners, reportTag, loadCsv, analytics_html, analytics, internal, JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, EXEC_KEY, RELATE_TYPE, XRAY_TEST_TYPE, ASSIGNEE, DEFAULT_STATUS, BRANCH, VERSION):
    from py.analytics import generate_advanced_analytics
    from py.html import prepare_html
    from py.functions import topFailedTC, format_df, hyperlink, prepare_failed_table, prepare_tags_table, topFailedReasons, prepare_module_graph, flatten_json, create_summary_pie, prepare_device_coverage, get_resources, prepare_custom_failure_graph, generateExcel
    import py.custom_reasons as cr
    import py.integrations.jira as jira
    truncated = True
    page = 1
    i = 0
    daysOlder = 0
    resources = []
    failurecategorytable, topfailedtc_table, tags_df_table, tags_base64 = "", "", "", ""

    if loadCsv == "":
        resources = get_resources(
            truncated, reportTag, page, daysOlder, resources, jobName, jobNumber, startDate, endDate, internal)

        if len(resources) > 0:
            # update tags - Pools no use
            rem_list = ['uxDuration', 'parameters', 'customFields',
                        'executionEngine', 'selectionCriteriaV2', 'artifacts', 'version',
                        'testExecutionId', 'retry', 'hidden', 'duration', 'tracking', 'uiOptions']
            vid_list = ['startTime', 'endTime', 'format', 'screen']
            platform_list = ['deviceId', 'triggerType',
                             'selectionCriteria', 'selectionCriteriaV2', 'customFields']
            for x in range(len(resources)):
                resources[x]['tags'] = [ele.replace("@", "") for ele in resources[x]["tags"] if not re.compile(
                    r"([0-9a-zA-Z-]){34}|"+tags_to_remove).match(ele)]
                resources[x] = {key: resources[x][key]
                                for key in resources[x] if key not in rem_list}
                if internal == "false":
                    if len(resources[x]['videos']) > 0:
                        for v in range(len(resources[x]['videos'][0])):
                            resources[x]['videos'][0] = {key: resources[x]['videos'][0][key]
                                                         for key in resources[x]['videos'][0] if key not in vid_list}
                if len(resources[x]['platforms']) > 0:
                    for v in range(len(resources[x]['platforms'][0])):
                        resources[x]['platforms'][0] = {key: resources[x]['platforms'][0][key]
                                                        for key in resources[x]['platforms'][0] if key not in platform_list}

            jsonDump = json.dumps(resources)
            resources = json.loads(jsonDump)
            print("Total executions: " + str(len(resources)))
            df = pandas.DataFrame([flatten_json(x)
                                  for x in resources])  # no use of pool
            df = format_df(df, jobName, jobNumber, automation_owners, internal)

            if internal == "true":
                df['reportURL'] = 'https://' + os.environ["cloudName"] + \
                    '.app.perfectomobile.com/reporting/test/' + \
                    df['id'].astype(str)

            # export df to csv
            xl_thread = threading.Thread(target=generateExcel, name="Downloader", args=[
                df])
            xl_thread.start()
        else:
            print("0 test executions")
            exit(0)

    else:
        df = pandas.read_csv(loadCsv, low_memory=False)

    total = df.shape[0]

    # replace failed in cleanException
    if not 'cleanException' in df.columns:
        failure_items, custom_failure_items, failedTable = "", "", ""
    else:
        # don't put the below two methods below hyperlink
        async def do_multiple_tasks():
            return await asyncio.gather(cr.get_failure_reason_category_table(df, total),
                                        topFailedTC(total, df[(df["Status"] == "FAILED")]))
        failurecategorytable, topfailedtc_table = asyncio.run(
            do_multiple_tasks())

        # clean error message
        df = df.reset_index()
        df["cleanException"] = df["cleanException"].fillna("")
        df['cleanException'] = df['cleanException'].str.replace(
            r'\s+failed$|^Step\:And\s+|^Step\:Then\s+|^Step\:Given\s+|^Step\:When\s+', '', regex=True).astype('str')
        df["message"] = df["message"].fillna("")
        df['message'] = df['message'].astype(
            'string').str.replace(r'Step.*\n|.*Error\:|.*Exception\:|\n.*|<[^>]+>|\.\.\.|.*\.xpath\:|.*\.cssSelector\:|.*intercepted\:(\s)+|(\s)+Other element.*|.*locate(\s)+element\:|.*reference\:|.*seconds\:|\: Timed out receiving message.*|\(\d*\.?\d.*seconds\)\.|\(com.qmetry.qaf.*', '', regex=True)

    report_filename = "PerfectReports.html"

    if (analytics.lower() == "true"):
        profile_thread = Process(target=generate_advanced_analytics, args=(df, client_logo, report_filename, os.path.dirname(
            os.path.abspath(__file__)), analytics_html, internal))
        profile_thread.start()
    else:
        profile_thread = Process()

    # based on descending end time, skiping if latest tests have passed
    issues_df = df.sort_values(by="endTime", ascending=False)
    latest = issues_df.groupby('name').first()
    remove_names = latest.drop(
        latest[latest["Status"] != "PASSED"].index)
    # Remove from issues_df the rows that matches the most recent passed tests names.
    issues_df = issues_df[~issues_df['name'].isin(
        remove_names.index.tolist())]
    #   gets latest test name that fails. Keep this here only.
    latest, mostfailedNames = jira.getFailedNames(issues_df)

    # hyperlinking name with report link
    df = hyperlink(df, "Result",
                   "reportURL", "name", "", False)

    passed = df[(df["Status"] == "PASSED")].shape[0]
    blocked = df[(df["Status"] == "BLOCKED")].shape[0]
    unknown = df[(df["Status"] == "UNKNOWN")].shape[0]
    failed_df = df[(df["Status"] == "FAILED")]
    failed = failed_df.shape[0]
    failed_blocked = df[(df["Status"] == "FAILED") |
                        (df["Status"] == "BLOCKED")]
    failed_blocked = failed_blocked[~failed_blocked['name'].str.contains(
        "Interactive session")]
    failed_df = failed_df[~failed_df['name'].str.contains(
        "Interactive session")]

    execution_summary = {}
    df.replace(np.nan, '', regex=True)  # important to display all modules
    # prepare Device coverage: don't async
    df, version_items = prepare_device_coverage(df)

    # bug analysis
    jira.bug_analysis(JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, EXEC_KEY,
                      RELATE_TYPE, XRAY_TEST_TYPE, ASSIGNEE, DEFAULT_STATUS, BRANCH, VERSION, latest, mostfailedNames, failed_df)

    failed_df["Custom Failure Reasons"] = failed_df["Custom Failure Reasons"].fillna(
        'Uncategorized')

    async def do_for_graphs():
        return await asyncio.gather(create_summary_pie(df, "", "Status", "count"),
                                    prepare_custom_failure_graph(
                                        failed_blocked),
                                    prepare_module_graph(df))
    execution_summary, custom_failure_items, tags_base64 = asyncio.run(
        do_for_graphs())
    lst = [failed_blocked]
    del failed_blocked
    del lst
    lst = [df]
    del df
    del lst
    if 'cleanException' in failed_df.columns:
        failed_df = failed_df.rename(
            columns={
                "cleanException": "Failed Steps",
                "message": "Failure Message",
            }
        )

        failed_df['Failed Steps'] = failed_df['Failed Steps'].astype(
            'string').str.replace(r'<[^>]+>|\.\.\.', '', regex=True)

        # create table for failures, pivot tags, prepare failed table
        async def do_if_failures():
            return await asyncio.gather(topFailedReasons(total, failed_df),
                                        prepare_tags_table(
                                            total, failed_df),
                                        prepare_failed_table(failed_df, internal))
        failure_items, tags_df_table, failedTable = asyncio.run(
            do_if_failures())
    lst = [failed_df]
    del failed_df
    del lst
    strTable = prepare_html(client_logo, criteria, total, passed, failed, unknown, blocked, execution_summary,
                            tags_base64, failure_items, custom_failure_items, version_items, tags_df_table, failedTable, topfailedtc_table, analytics_html, failurecategorytable)
    strTable = strTable.replace("<thead>", "<thead class='stuckHead'>")
    if profile_thread.is_alive():
        profile_thread.join()
    with open(report_filename, 'w') as f:
        f.write(strTable)
    file_path = os.path.join(os.getcwd(), report_filename)
    if (ms_webhook != ""):
        from py.integrations.teams import send_teams_message
        send_teams_message(ms_webhook, direct_link, file_path, criteria)
    print(
        "".join(["Report: file://", file_path])
    )
    if xl_thread.is_alive():
        xl_thread.join()


def main():
    tags_to_remove = ""
    automation_owners = ""
    from py.functions import get_key_details,  get_report_details
    try:
        #     """fix Python SSL CERTIFICATE_VERIFY_FAILED"""
        if not os.environ.get("PYTHONHTTPSVERIFY", "") and getattr(
            ssl, "_create_unverified_context", None
        ):
            ssl._create_default_https_context = ssl._create_unverified_context
        parser = ArgumentParser(
            description="PerfectReports")
        parser.add_argument(
            "-c",
            "--cloud_name",
            metavar="cloud_name",
            help="Perfecto cloud name. (E.g. demo) or add it as a cloudName environment variable",
            nargs="?",
        )
        parser.add_argument(
            "-s",
            "--security_token",
            metavar="security_token",
            type=str,
            help="Perfecto Security Token/ Pass your Perfecto's username and password in user:password format  or add it as a securityToken environment variable",
            nargs="?",
        )
        parser.add_argument(
            "-r",
            "--report",
            type=str,
            metavar="prepares custom report",
            help="creates a custom report.",
            nargs="?",
        )
        parser.add_argument(
            "-l",
            "--logo",
            type=str,
            metavar="customer logo link",
            help="Customer logo link",
            nargs="?",
        )
        parser.add_argument(
            "-remove-tags",
            "--tags_to_remove",
            type=str,
            metavar="Remove Tags",
            help="Tags that needs to be removed from the results",
            nargs="?",
        )
        parser.add_argument(
            "-automation-owners",
            "--automation_owners",
            type=str,
            metavar="Automation Owners",
            help="JSON format of Automation Owners and modules",
            nargs="?",
        )
        parser.add_argument(
            "-t",
            "--teams",
            type=str,
            metavar="MS Teams integration",
            help="Provide additional details for MS teams integration",
            nargs="?",
        )
        parser.add_argument(
            "-x",
            "--xray",
            type=str,
            metavar="Xray integration",
            help="Provide additional details for Xray integration",
            nargs="?",
        )
        args = vars(parser.parse_args())
        try:
            if not args["cloud_name"]:
                print("Loading cloudName: " +
                      os.environ["cloudName"] + " from environment variable.")
            else:
                os.environ["cloudName"] = args["cloud_name"]
        except Exception:
            if not args["cloud_name"]:
                parser.error(
                    "cloud_name parameter is empty. Either Pass the argument -c followed by cloud_name"
                )
                exit
            os.environ["cloudName"] = args["cloud_name"]
        try:
            if not args["security_token"]:
                print("Loading securityToken: " +
                      os.environ["securityToken"] + " from environment variable.")
            else:
                os.environ["securityToken"] = args["security_token"]
        except Exception:
            if not args["security_token"]:
                parser.error(
                    "security_token parameter is empty. Pass the argument -c followed by cloud_name"
                )
            os.environ["securityToken"] = args["security_token"]
        if args["logo"]:
            if str("www.").lower() not in str(args["logo"]).lower():
                raise Exception(
                    "Kindly provide valid client website url. Sample format: www.perfecto.io"
                )
            client_logo = args["logo"]
        else:
            client_logo = "https://www.perforce.com/sites/default/themes/custom/perforce/logo.svg"

        if args["tags_to_remove"]:
            tags_to_remove = args["tags_to_remove"]

        if args["automation_owners"]:
            automation_owners = args["automation_owners"]

        try:
            global criteria
            global jobNumber
            global jobName
            global startDate
            global endDate
            global ms_webhook
            global internal
            internal = "false"
            ms_webhook = ""
            loadCsv = ""
            jobName = ""
            jobNumber = ""
            startDate = ""
            endDate = ""
            temp = ""
            criteria = ""
            reportTag = ""
            analytics_html = "Analytics.html"
            analytics = "true"
            report = args["report"]
            report_array = report.split("|")
            JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, EXEC_KEY, RELATE_TYPE, XRAY_TEST_TYPE, ASSIGNEE, VERSION, BRANCH = "", "", "", "", "", "", "", "", "", ""
            DEFAULT_STATUS = "QA Review"
            for item in report_array:
                if "report" in item:
                    report, criteria = get_report_details(
                        item, temp, "report", criteria
                    )
                if "jobName" in item:
                    jobName, criteria = get_report_details(
                        item, temp, "jobName", criteria
                    )
                if "jobNumber" in item:
                    jobNumber, criteria = get_report_details(
                        item, temp, "jobNumber", criteria
                    )
                if "startDate" in item:
                    startDate, criteria = get_report_details(
                        item, temp, "startDate", criteria
                    )
                if "endDate" in item:
                    endDate, criteria = get_report_details(
                        item, temp, "endDate", criteria
                    )
                if "reportTag" in item:
                    reportTag, criteria = get_report_details(
                        item, temp, "reportTag", criteria
                    )
                if "loadCsv" in item:
                    loadCsv, criteria = get_report_details(
                        item, temp, "loadCsv", criteria
                    )
                if "analytics" in item:
                    analytics, criteria = get_report_details(
                        item, temp, "analytics", criteria
                    )
                if "internal" in item:
                    internal, criteria = get_report_details(
                        item, temp, "internal", criteria
                    )
        except Exception as e:
            raise Exception(
                "Verify parameters of report, split them by | seperator. Exception: " +
                str(e)
            )
            sys.exit(-1)
        if args["teams"]:
            try:
                global direct_link
                direct_link, temp = "", ""
                teams = args["teams"]
                team_array = teams.split("|")
                for item in team_array:
                    if "webhook" in item:
                        ms_webhook = get_key_details(item, temp, "webhook")
                    if "direct_link" in item:
                        direct_link = get_key_details(
                            item, temp, "direct_link")
            except Exception as e:
                raise Exception(
                    "Verify parameters of teams, split them by | seperator. Exception: " +
                    str(e)
                )
        if args["xray"]:
            try:
                xray = args["xray"]
                xray_array = xray.split("|")
                for item in xray_array:
                    if "JIRA_BASE_URL" in item:
                        JIRA_BASE_URL = get_key_details(
                            item, temp, "JIRA_BASE_URL")
                    if "JIRA_API_TOKEN" in item:
                        JIRA_API_TOKEN = get_key_details(
                            item, temp, "JIRA_API_TOKEN")
                    if "JIRA_USERNAME" in item:
                        JIRA_USERNAME = get_key_details(
                            item, temp, "JIRA_USERNAME")
                    if "PROJECT_KEY" in item:
                        PROJECT_KEY = get_key_details(
                            item, temp, "PROJECT_KEY")
                    if "EXEC_KEY" in item:
                        EXEC_KEY = get_key_details(
                            item, temp, "EXEC_KEY")
                        if (len(EXEC_KEY) > 0):
                            EXEC_KEY = EXEC_KEY.replace("@", "")
                    # Check Available Link Types using rest/api/2/issueLinkType & copy the appropriate link of outward contains the word relates to.
                    if "RELATE_TYPE" in item:
                        RELATE_TYPE = get_key_details(
                            item, temp, "RELATE_TYPE")
                    if "XRAY_TEST_TYPE" in item:
                        XRAY_TEST_TYPE = get_key_details(
                            item, temp, "XRAY_TEST_TYPE")
                    if "ASSIGNEE" in item:
                        ASSIGNEE = get_key_details(
                            item, temp, "ASSIGNEE")
                    if "DEFAULT_STATUS" in item:
                        DEFAULT_STATUS = get_key_details(
                            item, temp, "DEFAULT_STATUS")
                    if "VERSION" in item:
                        VERSION = get_key_details(
                            item, temp, "VERSION")
                    if "BRANCH" in item:
                        BRANCH = get_key_details(
                            item, temp, "BRANCH")
            except Exception as e:
                raise Exception(
                    "Verify parameters of teams, split them by | seperator. Exception: " +
                    str(e)
                )
        filelist = glob.glob(os.path.join("*.html"))
        for f in filelist:
            os.remove(f)
        if jobName:
            criteria += "JOB: " + jobName.replace(";", "; ").upper() + ", "
        if jobNumber != "":
            criteria += "JOB#: " + \
                jobNumber.replace(";", "; ") + " "
        if reportTag != "":
            criteria += "TAG:" + str(reportTag).upper() + ", "
        if startDate != "":
            if "-" not in startDate:
                criteria += "START: " + str(datetime.datetime.strptime(str(datetime.datetime.fromtimestamp(int(int(startDate) / 1000))), "%Y-%m-%d %H:%M:%S")) + \
                    ", END: " + \
                    str(datetime.datetime.strptime(str(datetime.datetime.fromtimestamp(
                        int(int(endDate) / 1000))), "%Y-%m-%d %H:%M:%S"))
            else:
                criteria += "START: " + startDate + ", END: " + endDate
        customReporter(client_logo, tags_to_remove, automation_owners,
                       reportTag, loadCsv, analytics_html, analytics, internal, JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, EXEC_KEY, RELATE_TYPE, XRAY_TEST_TYPE, ASSIGNEE, DEFAULT_STATUS, BRANCH, VERSION)

    except Exception as e:
        raise Exception("Oops!", e)


if __name__ == "__main__":
    freeze_support()
    start_time = datetime.datetime.now().replace(microsecond=0)
    main()
    end = datetime.datetime.now().replace(microsecond=0)
    print("Total Time taken:" + str(end - start_time))
    sys.exit()




def prepare_html(client_logo, criteria, total, passed, failed, unknown, blocked, execution_summary,
                 tags_base64, failure_items, custom_failure_items, version_items, tags_df_table, failedTable, topfailedtc_table, analytics_html, failurecategorytable):
    html = str('''
<!DOCTYPE html>
<script src="http://code.jquery.com/jquery-1.11.0.min.js"></script>
<META http-equiv="Content-Type" content="text/html; charset=UTF-8"> 
<html>
<header>
    <link href="https://cdn.datatables.net/1.11.5/css/jquery.dataTables.min.css" rel="stylesheet">
</header>
<head>
    <style>
    
        html {
            display:inline-grid;
        }
        
        body {
            font-family: sans-serif;
            font-size: 11px;
            margin: 0px;
            background: #f0f7f7bf;
        }

        .option2 {
            text-indent: 15px;
            display: inline-block;
            padding-right: 28px;
            text-align: right;
            margin: -4px;
            border-radius: 1em;
        }

        .stuckHead {
            position: sticky;
            top: 0;
            z-index: 1;
        }

        #tagsheading {
            border: 2px solid #90cfe27d;
        }
        table,
        th,
        td {
            border: 0px solid black;
            border-collapse: collapse;
            font-size: 11px;
        }

        .options2 {
            opacity: .8;
        }

        table.center {
            margin-left: auto;
            margin-right: auto;
        }

        th,
        td {
            padding: 2px;
            font-weight: 100;
        }

        th {
            text-align: center;
            box-sizing: border-box;
            border: 2px solid #366e9936;
            font-size: 12px;
            background: #5e90d8;
            font-weight: 600;
            color: white;
            text-align: center;
            line-height: 150%;
            vertical-align: middle;
        }

        td {
            box-sizing: border-box;
            border: 2px solid #366e9936;
            position: relative;
            color: #200302;
            background: #ffffff;
            vertical-align: middle;
            font-size: 11px;
            line-height: 140%;
            text-align: left;
        }

        img {
            display: inline-flex;
            height: 100%;
        }

        #summary {
            text-align: center;
            background: #f0ffff;
        }

        ::-webkit-scrollbar {
            -webkit-appearance: none;
            width: 5px;
            height: 5px;
        }

        ::-webkit-scrollbar-thumb {
            border-radius: 2px;
            background-color: rgba(0, 0, 0, .5);
            box-shadow: 0 0 1px rgba(255, 255, 255, .5);
        }

        #piechart,
        #browserChart {
            background-color: rgba(238, 240, 223, 0.68);
            display: block;
            width: 300px;
            height: 240px;
            margin: 5px;
            box-shadow: 0 0 60px rgb(80 130 50 / 33%);
        }

        .dataTables_wrapper .dataTables_filter input {
            padding: 3px;
        }
        
        #table tr td:nth-child(1), table.dataTable th:nth-child(1) {
          width: 50px;
          max-width: 50px;
        }
        #table tr td:nth-child(2), table.dataTable th:nth-child(2) {
            width: 80px;
            max-width: 80px;
        }
        
        #table tr td:nth-child(3), table.dataTable th:nth-child(3) {
            width:450px;
            max-width:450px;
        }

        #table tr td:nth-child(4), table.dataTable th:nth-child(4){
            width: 90px;
            max-width: 90px;
        } 
        
        #table tr td:nth-child(5), table.dataTable th:nth-child(5){
            max-width: 100px;
            width: 100px;
        }
        
        #table tr td:nth-child(6), table.dataTable th:nth-child(6) {
          max-width: 190px;
          width:190px;
        }

        table.dataTable th:nth-child(7) {
          width:35px;
          max-width:35px;
        }
      
        table.dataTable thead th, table.dataTable thead td {
          padding: 8px 0px;
        }
      
        #images {
            background: linear-gradient(to right, #ffb35a70, #E3F3FE, #E3F3FE, #aacfe8 90.33%, #ffb35a70);
            box-shadow: 0 1px 6px rgba(0, 0, 0, 0.12), 0 1px 4px rgba(0, 0, 0, 0.24);
            height: 35px;
            margin: auto;
            display: flex;
            justify-content: center;
            padding: 2px 1px 2px 10px;
        }

        #clientlogo {
            width: 100px;
            padding-left: 20px;
        }

        #perfecto {
            width: 100px;
        }

        #criteria-legend {
            box-shadow: 0 1px 6px rgba(0, 0, 0, 0.12), 0 1px 4px rgba(0, 0, 0, 0.24);
            background: #78d9e891;
            background-image: linear-gradient(#78d9e891, #d9f0f491, #ffffffc2);
            padding: 5px;
            display: flex;
            justify-content: center;
            padding-top: 9px;
        }

        #summary-legend {
            box-shadow: 0 1px 6px rgba(0, 0, 0, 0.12), 0 1px 4px rgba(0, 0, 0, 0.24);
            background: #ffb35a70;
            padding: 5px;
        }

        #table, #tableDiv {
            box-shadow: 0 10px 10px rgb(80 130 50 / 33%);
            overflow-y: scroll;
            word-wrap: break-word;
            margin-top: 0px;
            max-height: 533px;
            width: 770px;
            border: 2px solid #90cfe27d;
        }

        #moduleheading {
            overflow: scroll;
            height: 210px;
            max-width: 400px;
        }

        #hie {
            display: flex;
            flex-wrap: wrap;
            flex-direction: column;
            align-content: center;
        }

        .option2 {
            font-size: small;
        }
        
        #vidlink {
            background-image: url(https://img.icons8.com/office/16/link.png);
            background-position: 2px 0px;
            background-repeat: no-repeat;
            height: 16px;
            width: 18px;
        }

        #tags, #category,
        #topfailedtests, #topfailedmessages {
            overflow-y: scroll;
            word-wrap: break-word;
            display: inline-block;
            border-collapse: collapse;
            table-layout: fixed;
            max-height: 128px;
            width: 100%;
        }

        #tags tbody tr th, #category tbody tr th,
        #topfailedtests tbody tr th, #topfailedtests tbody tr th, #topfailedtests tbody tr,
        #topfailedmessages tbody tr th, #topfailedmessages tbody tr th, #topfailedmessages tbody tr {
            text-align: left;
            background: #fffaf2;
            color: black;
            font-weight: 100;
            font-size: 10px;
        }

        #category tbody tr th {
            width: 100%;
        }
        #topfailedtests tbody tr th,
        #topfailedmessages tbody tr th {
            width: 200px !important;
        }

        #topfailedtests tbody tr td:nth-child(1),
        #topfailedmessages tbody tr td:nth-child(1) {
            min-width: 385px;
            max-width: 385px; 
            width: 100%;
        }
        
        #tags tbody tr th:nth-child(3) {
            width: 100%;
        }

        #category tbody tr th:nth-child(3){
            min-width: 220px;
        }
        #tags thead tr:nth-child(1), #category thead tr:nth-child(1) {
            display: none;
        }
        
        #tags thead tr:nth-child(2) th:nth-child(4):before, #category thead tr:nth-child(2) th:nth-child(4):before  {
            content: "Total";
        }
        
        #tags thead tr:nth-child(2) th:nth-child(5):before, #category thead tr:nth-child(2) th:nth-child(5):before{
            content: "✓%↑";
        }

        #slide {
            display: inline-flex;
            background: #f8ffff;
            flex-direction: row;
            overflow: scroll;
            align-items: self-start;
            justify-content: flex-start;
        }

        a {
            color: #04192d;
        }

        #heading {
            font-size: small;
            display: flex;
            flex-direction: row;
            align-items: center;
            justify-content: center;
            font-weight: bold;
            width: 100%;
            padding: 5px 0;
            text-align: center;
            cursor: pointer;
            color: white;
            background: #002e5dd1;
            height: 15px;
        }

        svg {
            overflow: scroll;
        }

        #box {
            box-shadow: 0 -4px 18px 0 rgba(60, 118, 191, 0.55);
            margin-right: 2px;
            margin-left: 2px;
            margin-top: 3px;
            border: 1px solid #89def77d;
        }

        #topfailbox {
            margin: 3px;
            max-height: 310px;
            width: 380px;
            overflow: scroll;
            display: inline-table;
        }
        
        #reporticon {
            padding-left: 20px;   
        }
    </style>
</head>
<title>PerfectReports</title>

<body>
    <div id="images">
        <img src='https://www.perfecto.io/sites/default/themes/custom/perfecto/logo.svg?height=57&width=200'
            alt="Perfecto" id="perfecto" class="center">
        <img src="''' + client_logo + '''" alt="logo" id="clientlogo" class="center">
        <a href="./''' + analytics_html + '''" target="_blank" ><img width="40" height="35" src="https://img.icons8.com/external-smashingstocks-thin-outline-color-smashing-stocks/67/external-Data-Analytics-industrial-production-smashingstocks-thin-outline-color-smashing-stocks-2.png" id="reporticon"
      alt="graph-report" /></a>
    </div>
    <div id="criteria-legend">
        <h2 class="option2" style="background-color:#ffd7aed4;">''' + str(criteria) + '''</h2>
        <h2 class="option2" style="background-color:lightgoldenrodyellow;">''' + str(total) + ''' \nTESTS</h2>
        <h2 class="option2" style="background-color:#33d63391;">''' + str(passed) + ''' \nPASSED</h2>
        <h2 class="option2" style="background-color:#ff5f41b5;">''' + str(failed) + ''' \nFAILED</h2>
        <h2 class="option2" style="background-color:#f2ebeb;">''' + str(unknown) + ''' \nUNKNOWNS</h2>
        <h2 class="option2" style="background-color:#ffa50091">''' + str(blocked) + ''' \nBLOCKED</h2>
    </div>
    <div id="summary">
        <div id="slide">
            <div id="box"><div id="heading">Overall Status</div>''' + execution_summary + '''</div></div>
            <div id="box"><div id="heading">Module-wise Status</div><div id="moduleheading">''' + tags_base64 + '''</div></div>
            ''' + version_items + ''' 
            <div id="box"><div id="heading">Top 5 Custom Failure Reasons</div><div>''' + custom_failure_items + '''</div></div>
        </div>
    </div>
    <div id="summary">
        <div id="slide">
            <div id="topfailbox">
                <div id="heading">Top 10</div>
                <div id="tagsheading">''' + failure_items + '''</div>
                <div id="tagsheading">''' + failurecategorytable + '''</div>
                <div id="tagsheading">''' + tags_df_table + '''</div>
                <div id="tagsheading">''' + topfailedtc_table + '''</div>
            </div>
            <div id="box">
                <div id="heading">Failed Tests Summary</div>
                <div id="tableDiv">''' + failedTable + '''</div>
            </div>
        </div>
    </div>
</body>
</script>
<script src="https://code.jquery.com/jquery-3.6.0.slim.min.js" integrity="sha256-u7e5khyithlIdTpu22PHhENmPcRdFiHRjhAuHcs05RI=" crossorigin="anonymous"></script>
<script type="text/javascript" src="https://cdn.datatables.net/1.11.5/js/jquery.dataTables.min.js"></script>
<script>
    $(document).ready( function () {
        $('#table').DataTable({
            paging: true,
            searching: true,
            responsive: true,
            "autoWidth": false,
            "pageLength": 4
        });
    });
</script>
</html>
        ''')
    return html



from datetime import datetime
import requests
from requests.auth import HTTPBasicAuth
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import json


def create_jira_session():
    # Headers
    headers = {
        'Accept': 'application/json',
        'Content-Type': 'application/json'
    }

    # Setup retry strategy
    retry_strategy = Retry(
        total=5,  # Total number of retries
        # Retry on these HTTP status codes
        status_forcelist=[429, 500, 502, 503, 504],
        # Retry for these HTTP methods
        allowed_methods=["HEAD", "GET", "OPTIONS"],
        backoff_factor=1  # Exponential backoff factor
    )

    # Setup the adapter with the retry strategy
    adapter = HTTPAdapter(max_retries=retry_strategy)

    # Create a session and mount the adapter
    session = requests.Session()
    session.mount("https://", adapter)
    session.mount("http://", adapter)
    return headers, session


def get_jira_issues(JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, JIRA_JQL_QUERY):
    headers, session = create_jira_session()
    URL = f'{JIRA_BASE_URL}/rest/api/2/search'
    print(URL)
    # Parameters
    max_results = 100
    start_at = 0
    total_issues = []
    is_last = False
    while not is_last:
        params = {
            'jql': JIRA_JQL_QUERY,
            'maxResults': max_results,
            'startAt': start_at
        }

    # Perform the request with retries and a timeout of 2 minutes
        try:
            response = session.get(
                URL,
                headers=headers,
                params=params,
                auth=HTTPBasicAuth(JIRA_USERNAME, JIRA_API_TOKEN),
                timeout=(120, 120)  # Connection timeout, Read timeout
            )

            # Check for successful response
            if response.status_code == 200:
                data = response.json()
                issues = data.get('issues', [])
                total_issues.extend(issues)
                start_at += max_results
                is_last = start_at >= data.get('total', 0)
            else:
                print("Failed to query JIRA URL: " + URL + ", " +
                      str(response.status_code) + " : " + str(response.text))
                break
        except requests.exceptions.RequestException as e:
            print("Request failed: " + str(e))
            break
    return total_issues


def create_jira_bug(JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, LABELS, ASSIGNEE, SUMMARY, DESCRIPTION):
    issue_key = ''
    headers, session = create_jira_session()
    URL = f'{JIRA_BASE_URL}/rest/api/2/issue'
    # Issue data
    issue_data = {
        "fields": {
            "project": {
                "key": PROJECT_KEY
            },
            "summary": SUMMARY,
            "description": DESCRIPTION,
            "issuetype": {
                "name": "Bug"
            },
            "assignee": {
                "accountId": ASSIGNEE
            },
            "priority": {
                "name": "Medium"
            },
            "labels": LABELS
        }
    }
    try:
        response = session.post(
            URL,
            headers=headers,
            data=json.dumps(issue_data),
            auth=HTTPBasicAuth(JIRA_USERNAME, JIRA_API_TOKEN),
            timeout=(120, 120)  # Connection timeout, Read timeout
        )

        # Check for successful response
        if response.status_code == 201:
            issue_key = response.json().get('key')
            print("Successfully created issue: " + issue_key)
        else:
            print("Failed to create JIRA bug. URL: " + URL + ", " +
                  str(response.status_code) + " : " + str(response.text))

    except requests.exceptions.RequestException as e:
        print("Request failed: " + str(e))
    return issue_key


def update_bug_links(JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, RELATE_TYPE, BUG_KEY, TEST_KEYS):
    headers, session = create_jira_session()
    URL = f'{JIRA_BASE_URL}/rest/api/2/issueLink'
    for TEST_KEY in TEST_KEYS:
        link_data = {
            "type": {
                "name": RELATE_TYPE
            },
            "inwardIssue": {
                "key": TEST_KEY
            },
            "outwardIssue": {
                "key": BUG_KEY
            }
        }

        try:
            response = session.post(
                URL,
                headers=headers,
                data=json.dumps(link_data),
                auth=HTTPBasicAuth(JIRA_USERNAME, JIRA_API_TOKEN),
                timeout=(120, 120)  # Connection timeout, Read timeout
            )

            if response.status_code == 201:
                print("Successfully linked bug: " +
                      BUG_KEY + " to test " + TEST_KEY)
            else:
                print("Failed to update JIRA bug. URL: " + URL + ", " +
                      str(response.status_code) + " : " + str(response.text))
        except requests.exceptions.RequestException as e:
            print("Request failed: " + str(e))


def delete_jira_bug(JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, ISSUE_KEY):
    headers, session = create_jira_session()
    URL = f'{JIRA_BASE_URL}/rest/api/2/issue/{ISSUE_KEY}'

    try:
        response = session.delete(
            URL,
            headers=headers,
            auth=HTTPBasicAuth(JIRA_USERNAME, JIRA_API_TOKEN),
            timeout=(120, 120)  # Connection timeout, Read timeout
        )

        # Check for successful response
        if response.status_code == 204:
            print("Successfully deleted issue: " + ISSUE_KEY)
        else:
            print("Failed to delete JIRA bug. URL: " + URL + ", " +
                  str(response.status_code) + " : " + str(response.text))
    except requests.exceptions.RequestException as e:
        print("Request failed: " + str(e))


def get_jira_labels(JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, ISSUE_KEY):
    labels = []
    headers, session = create_jira_session()
    URL = f'{JIRA_BASE_URL}/rest/api/2/issue/{ISSUE_KEY}'

    try:
        response = session.get(
            URL,
            headers=headers,
            auth=HTTPBasicAuth(JIRA_USERNAME, JIRA_API_TOKEN),
            timeout=(120, 120)  # Connection timeout, Read timeout
        )

        # Check for successful response
        if response.status_code == 200:
            issue_data = response.json()
            labels = issue_data.get('fields', {}).get('labels', [])
            print("Labels for issue " + str(ISSUE_KEY) + ": " + ", ".join(labels))
        else:
            print("Failed to retrieve issue. URL: " + URL + ", " +
                  str(response.status_code) + " : " + str(response.text))
    except requests.exceptions.RequestException as e:
        print("Request failed: " + str(e))
    return labels


def jira_bug(row, JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, EXEC_KEY, RELATE_TYPE, XRAY_TEST_TYPE, ASSIGNEE, DEFAULT_STATUS, BUG_LABEL):
    SUMMARIES = row['name']
    if (JIRA_BASE_URL != ""):
        if (JIRA_API_TOKEN != ""):
            if (PROJECT_KEY != ""):
                if (EXEC_KEY != ""):
                    LABELS = [BUG_LABEL, "automation", "perfecto"]
                    SUMMARY = EXEC_KEY + ":" + row['message']
                    JQL_BUG_SUMMARY = "project = '" + PROJECT_KEY + \
                        "' AND type = Bug AND status != '" + DEFAULT_STATUS + "' AND labels = '" + \
                        BUG_LABEL + "' AND summary ~ '" + \
                        escapeSpecialChars(SUMMARY) + "'"
                    return raise_bug(JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, EXEC_KEY, RELATE_TYPE,
                                     ASSIGNEE, LABELS, SUMMARY, XRAY_TEST_TYPE, SUMMARIES, JQL_BUG_SUMMARY)
                else:
                    return 'No Execution key found'
            else:
                return 'No Project key found'
        else:
            return 'No JIRA Token!'
    else:
        return 'No JIRA URL!'


def escapeSpecialChars(summary):
    return summary.replace(' -', r' \\-').replace("'", r"''").replace('"', r'\\"').replace('@', r'\\@').replace('[', r'\\[').replace(']', r'\\]').replace('*', r'\\*')


def create_jira_link(jira_key, JIRA_BASE_URL):
    return f'<a href="{JIRA_BASE_URL}/browse/{jira_key}">{jira_key}</a>'


def get_test_keys(JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, XRAY_TEST_TYPE, SUMMARIES):
    TEST_KEYS = []
    for SUMMARY in SUMMARIES:
        # escape sp chars to avoid jql issues
        SUMMARY = escapeSpecialChars(SUMMARY)
        JQL_JUST_SUMMARY = "project = '" + PROJECT_KEY + \
            "' AND type = '" + XRAY_TEST_TYPE + "' AND summary ~ '" + SUMMARY + "'"
        total_issues = get_jira_issues(
            JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, JQL_JUST_SUMMARY)
        # list xray test
        print("Found " + str(len(total_issues)) +
              " issue with summary: " + SUMMARY)
        if (len(total_issues)) > 0:
            for issue in total_issues:
                TEST_KEY = issue.get("key", "")
                if (TEST_KEY != ""):
                    TEST_KEYS.append(TEST_KEY)
    return TEST_KEYS


def getTestExecutionsList():
    # TODO: create xray header
    url = "https://xray.cloud.getxray.app/api/v2/graphql"
    start = 0
    limit = 1
    total = 0
    test_list = []
    test_list_str = ""
    EXEC_KEY = "XS-41"
    while True:
        payload = json.dumps({
            "query": "{step1: getTestExecutions(jql:\"key=" + EXEC_KEY + "\", limit: 1) {    total    start    limit    results {      issueId      jira(fields: [\"key\"])      projectId      tests(start: " + str(start) + ", limit: " + str(limit) + ") {        total        start        limit        results { issueId         jira(fields: [\"key\"]) testType{ name }        }      }    }  }}",
            "variables": {}
        })
        headers = {

            'Authorization': 'Bearer <TOKEN>',
            'Content-Type': 'application/json'
        }
        try:
            response = requests.request(
                "POST", url, headers=headers, data=payload)
            if (response.status_code == 200):
                mapper = response.json()
                step = mapper.get("data", {}).get("step1", {})
                tests = step.get("results", [{}])[0].get("tests", {})
                results = tests.get("results", [])

                for execution in results:
                    test_type = execution.get("testType", {}).get("name")
                    if test_type:
                        if test_type.lower() == "cucumber":
                            key = execution.get("jira", {}).get("key", "")
                            if key:
                                test_list.append(key)

                # Convert test_list to a comma-separated string
                test_list_str = ",".join(test_list)

                # Get total and increment start by limit (assuming limit is defined somewhere)
                total = tests.get("total", 0)
                # Increment start for the next iteration
                start += limit
                print("Total:", total)
                print("Start:", start)
                # Check if start is less than total for the next iteration
                if start == total:
                    break
            else:
                print("Xray Get Executions Error:", response.reason)
        except Exception as e:
            print(e)
    return test_list_str


def raise_bug(JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, EXEC_KEY, RELATE_TYPE, ASSIGNEE, LABELS, SUMMARY, XRAY_TEST_TYPE, SUMMARIES, JQL_BUG_SUMMARY):
    # get test keys
    TEST_KEYS = get_test_keys(JIRA_BASE_URL, JIRA_API_TOKEN,
                              JIRA_USERNAME, PROJECT_KEY, XRAY_TEST_TYPE, SUMMARIES)

    if (len(TEST_KEYS) > 0):
        # Checks for existing issues with different statuses except To Do and don't create a bug if exists.
        total_issues = get_jira_issues(
            JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, JQL_BUG_SUMMARY)
        print("Found " + str(len(total_issues)) +
              " issues with JQL: " + JQL_BUG_SUMMARY)
        if (len(total_issues)) == 0:
            # fetch labels of xray test
            labels = get_jira_labels(JIRA_BASE_URL, JIRA_API_TOKEN,
                                     JIRA_USERNAME, TEST_KEYS[0])
            # create bug with that labels
            labels = labels + LABELS + TEST_KEYS
            description = "Impacted Xray Tests: \n" + "\n".join(TEST_KEYS)
            BUG_KEY = create_jira_bug(JIRA_BASE_URL, JIRA_API_TOKEN,
                                      JIRA_USERNAME, PROJECT_KEY, labels, ASSIGNEE, SUMMARY, description)
            if (len(BUG_KEY)) > 0:
                if (EXEC_KEY != ""):
                    TEST_KEYS.append(EXEC_KEY)
                update_bug_links(JIRA_BASE_URL, JIRA_API_TOKEN,
                                 JIRA_USERNAME, RELATE_TYPE, BUG_KEY, TEST_KEYS)
                return BUG_KEY
            else:
                return 'NA'
        else:
            return ",".join([item.get('key') for item in total_issues if item.get('key') is not None])
    return 'JIRA Summary not found!'


def getFailedNames(issues_df):
    mostfailedNames = []
    for name in issues_df['name'].unique():
        trends = (issues_df[issues_df["name"] == name]
                  ['Status'].value_counts(normalize=True) * 100).round(2)
        if ('FAILED' in trends.index):
            mostfailedNames.append(name)
    latest = issues_df
    lst = [issues_df]
    del issues_df
    del lst
    return latest, mostfailedNames


def bug_analysis(JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, EXEC_KEY, RELATE_TYPE, XRAY_TEST_TYPE, ASSIGNEE, DEFAULT_STATUS, BRANCH, VERSION, latest, mostfailedNames, failed_df):
    # drops duplicate test names
    latest.index = latest['name']
    latest = latest[latest.index.isin(mostfailedNames)]
    latest = latest.drop_duplicates('name', keep='first')
    if (len(latest)) > 0:
        latest = latest.groupby('message')['name'].apply(
            lambda x: list(x.index)).reset_index()

    BUG_LABEL = EXEC_KEY + "-bug"
    remove_old_bugs(JIRA_BASE_URL, JIRA_API_TOKEN,
                    JIRA_USERNAME, PROJECT_KEY, BUG_LABEL, DEFAULT_STATUS)

    if (len(failed_df) > 0):
        latest['JIRA KEY'] = latest.apply(
            lambda row: jira_bug(row, JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, EXEC_KEY, RELATE_TYPE, XRAY_TEST_TYPE, ASSIGNEE, DEFAULT_STATUS, BUG_LABEL), axis=1)
        print(latest)
        # Apply the function to the 'JIRA KEY' column
        latest['JIRA KEY'] = latest['JIRA KEY'].apply(
            lambda x: create_jira_link(x, JIRA_BASE_URL))
        latest['name'] = latest['name'].apply(lambda x: '<br>'.join(x))
    else:
        latest['JIRA KEY'] = ''
        latest['message'] = ''
        latest['name'] = ''
    latest.rename(columns={'message': 'Actual Error',
                           'name': 'Impacted Xray Test Summaries', 'cleanException': 'Failed Step'}, inplace=True)
    latest = latest[['JIRA KEY', 'Actual Error',
                     'Impacted Xray Test Summaries']]
    branchVersionInfo = ""
    if len(BRANCH) > 0:
        branchVersionInfo = """<table border="1" class="dataframe center table table-hover"><tbody><tr>
                <th>TeraScript-Branch</th><th>""" + BRANCH + """</th><td>Java-Version</td><td>""" + VERSION + """ </td > </tr> </tbody></table><br>"""

    from datetime import datetime
    style = """
            <style >
                body {
                    font-family: sans-serif;
                    font-size: 11px;
                    margin: 0px;
                    background:  #f0ffff;
                }
                table {
                    font-family: sans-serif;
                    font-size: 11px;
                    border-collapse: collapse;
                }
                th, td {
                    padding: 5px;
                    text-align: left;
                    border: 2.6px solid #16151778;
                }
                td {
                    background-color: seashell;
                    color: darkslategrey;
                }
                th {
                    background-color: #bc655e;
                    color: white;
                    text-align: center;
                }
                tr:nth-child(even) {
                    background-color:  #f9f9f9;
                }
                tr:hover {
                    background-color:  #f1f1f1;
                }
                .center {
                    margin-left: auto;
                    margin-right: auto;
                }
                #perfecto {
                    width: 100px;
                }
                img {
                    display: inline-flex;
                    height: 100% ;
                }
                #images {
                    background: linear-gradient(to right,  #ffb35a70, #E3F3FE, #E3F3FE, #aacfe8 90.33%, #ffb35a70);
                    box-shadow: 0 1px 6px rgba(0, 0, 0, 0.12), 0 1px 4px rgba(0, 0, 0, 0.24);
                    height: 35px;
                    margin: auto;
                    display: flex;
                    justify-content: center;
                    padding: 2px 1px 2px 10px;
                    align-items: center;
                }
                #criteria-legend {
                    background:  #78d9e891;
                    background-image: linear-gradient(#78d9e891, #d9f0f491, #ffffffc2);
                    display: flex;
                    justify-content: center;
                    padding-top: 9px;
                }
            </style>
            <body>
            <div id = "images">
                <img src = "https://www.perfecto.io/sites/default/themes/custom/perfecto/logo.svg?height=57&amp;width=200" alt = "Perfecto" id = "perfecto" class = "center">
                <h3 style="border: 4.6px solid #3d908178;padding: 10px;border-radius: 10px;"> """ + create_jira_link(EXEC_KEY, JIRA_BASE_URL) + """ BUG REPORT: """ + datetime.now().strftime('%Y-%m-%d %H:%M:%S') + """ </h3>
                <img src = "https://www.medanswering.com/images/MAS_web.png" alt = "logo" id = "clientlogo" class = "center">
            </div> <div id = "criteria-legend" > </div > <br>
            """ + branchVersionInfo
    html_string = style + latest.to_html(classes='center table table-hover',  # Adding CSS classes
                                         border=1,  # Adding border
                                         index=False,  # Hiding the index
                                         escape=False,
                                         # Formatting the 'JIRA#' column
                                         formatters={
                                             'JIRA#': '{:03d}'.format}
                                         )
    lst = [latest]
    del latest
    del lst
    from pathlib import Path
    BUG_REPORT_FILENAME = str(Path.cwd()) + '/Bug_Report.html'
    import os
    try:
        if os.path.exists(BUG_REPORT_FILENAME):
            os.unlink(BUG_REPORT_FILENAME)
    except Exception as e:
        print("Error: " + str(e))
    with open(BUG_REPORT_FILENAME, 'w') as f:
        f.write(html_string)
    print("\n================================\nBug report: " +
          str(BUG_REPORT_FILENAME))


def remove_old_bugs(JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, PROJECT_KEY, BUG_LABEL, DEFAULT_STATUS):
    if (JIRA_BASE_URL != ""):
        JQL_BUG_LABEL = 'project = ' + PROJECT_KEY + \
            ' AND type = Bug AND status = "' + DEFAULT_STATUS + '" AND labels = ' + BUG_LABEL
        total_issues = get_jira_issues(
            JIRA_BASE_URL, JIRA_API_TOKEN, JIRA_USERNAME, JQL_BUG_LABEL)
        # Remove old bugs
        print("Found " + str(len(total_issues)) +
              " issues with JQL: " + JQL_BUG_LABEL)
        if (len(total_issues)) > 0:
            for issue in total_issues:
                delete_jira_bug(JIRA_BASE_URL, JIRA_API_TOKEN,
                                JIRA_USERNAME, issue["key"])
    else:
        print("No JIRA URL specified!")




def read_report(file):
    from bs4 import BeautifulSoup
    status = []
    topfailedmessages = []
    # read report
    with open(file) as inf:
        txt = inf.read()
        soup = BeautifulSoup(txt, "html.parser")
        soup.prettify(formatter="minimal")
    # fetch status
    for div in soup.findAll('h2', {'class': 'option2'}):
        text = div.text
        if ':' not in text:
            status.append(text.replace("\n", ' ').replace("  ", " ").strip())
    # fetch top 3 failure messages
    i = 0
    if len(soup.find_all('table', {'id': 'topfailedmessages'})) > 0:
        for tr in soup.find_all('table', {'id': 'topfailedmessages'})[0].findAll('tr'):
            text = tr.text
            if 'Failure' not in text:
                topfailedmessages.append(
                    text[:-1].strip().replace('\n', '::: ').replace(text[:1], '').replace(text[:1], ''))
                i = i+1
                if (i == 3):
                    break
    return status, topfailedmessages


def send_teams_message(ms_webhook, direct_link, result_file, criteria):
    import pymsteams
    import os
    fails = []
    # get status and top 3 failure messages
    status, topfailedmessages = read_report(result_file)
    # card title
    myTeamsMessage = pymsteams.connectorcard(ms_webhook)
    if (criteria.startswith("START")):
        # get date
        criteria = criteria.split(",")[0].split("START:")[1].strip()
    myTeamsMessage.text(
        "".join(["**", os.environ["cloudName"].upper(), " Automation Reports Summary of ", criteria.strip(), "** "]))
    # prepare failure messages
    for item in topfailedmessages:
        fails.append(
            "".join(["- ", item.split(":::")[0].strip(), " -> _", item.split(":::")[2].strip(), "_", " \r"]))

    myMessageSection = pymsteams.cardsection()
    myMessageSection.addFact("Status:", ', '.join(status))
    if len(fails) > 0:
        myMessageSection.addFact("Top 3 Failure insights:", "".join(fails))
    myTeamsMessage.addSection(myMessageSection)
    # create direct azure report link if needed.
    if direct_link != "":
        myTeamsMessage.addLinkButton(
            "View Report", direct_link.replace(r' ', '%20'))
    myTeamsMessage.send()



import pandas
import requests
import os
import json
from functions import *

def get_test_execution_commands(test_id):
    url = "https://" + os.environ["cloudName"] + \
        ".app.perfectomobile.com"
    api_url = url + "/export/api/v3/test-executions/" + test_id + "/commands"
    payload = {}
    headers = {
        'Perfecto-Authorization': os.environ['securityToken'],
        'Content-Type': 'application/json',
        'Perfecto-TenantId': os.environ["cloudName"] + '-perfectomobile-com'
    }

    response = requests.request("GET", api_url, headers=headers, data=payload)
    return response.content

def get_test_steps_info(test_id):
    result = get_test_execution_commands(test_id)
    result = json.loads(result)

    try:
        resultList = result["resources"]
    except TypeError:
        print(result)

    if (len(resultList) > 0):
        df = pandas.DataFrame([flatten_json(x) for x in resultList])
        # last screenshot API URL
        df.loc[:, df.columns.str.endswith(
            "screenshots/0")].iloc[-1:, -1:].values[0][0]
        steps = "\n".join(df['name'].to_list())
        return steps


def get_devices_list():
    os.environ["cloudName"]= 'mas'
    os.environ["securityToken"] = os.environ["MAS_TOKEN"]
    url = "https://" +  os.environ["cloudName"]  + \
        ".app.perfectomobile.com/api/v1/device-management/devices"
    payload = json.dumps({
        "device": {}
    })
    headers = {
        'Perfecto-Authorization': os.environ['securityToken'],
        'Content-Type': 'application/json'
    }
    response = requests.request("POST", url, headers=headers, data=payload)
    try:
        response = json.loads(response.text)
        resultList = response["handsets"]['handset']
    except TypeError:
        print(response)

    if (len(resultList) > 0):
        df = pandas.DataFrame([flatten_json(x) for x in resultList])
        # df = df[['model', 'available', 'osVersion', 'deviceId', 'manufacturer', 'resolution']]
        # print(df)
        # return df
    return json.loads(json.dumps(response))

# print(get_test_steps_info(test_id))




from base64 import b64encode
import pandas
import plotly.graph_objects as go
from statsmodels.tsa.api import ExponentialSmoothing
from datetime import timedelta, datetime
import pandas as pd
import numpy as np

df = pandas.read_csv(
    '/Users/genesis.thomas/workspace/python/generic/PerfectoCustomReport/src/perfecto/output1.csv', low_memory=False)

df = df[(df["Status"] == "FAILED")]


def forecast_failures(df):

    if len(df.week.unique()) > 1:
        seasonal_periods = 5
        predict_days = 5
    else:
        seasonal_periods = 2
        predict_days = 2
    df["startHour"] = pandas.to_datetime(
        pandas.to_datetime(df["startTime"], format="%Y-%m-%d %H:%M:%S")
        .dt.to_period("H")
        .astype(str)
    )
    
    df_hour = df['startHour'].value_counts().rename_axis('Hour').reset_index(
        name='counts').sort_values(by=['Hour'], ascending=True)
    df_hour = df_hour.set_index(["Hour"])
    df_hour = df_hour.resample('H').mean().replace(np.nan, 0)

    start_time = str(df_hour.index[-1])
    future_time = str(df_hour.index[-1] + timedelta(days=predict_days))
    futureDate = str((datetime.strptime(
        future_time, "%Y-%m-%d %H:%M:%S") + timedelta(days=1)).date())
    predict_df = pd.DataFrame(
        {'Hour': pd.date_range(
            start_time.split(" ")[0], futureDate, freq='1H')}
    )
    predict_df = predict_df.set_index(["Hour"])
    predict_df['counts'] = 0
    y_hat_avg = predict_df.copy()
    y_hat_avg['forecast'] = ExponentialSmoothing(
        df_hour['counts'], seasonal_periods=seasonal_periods, trend='add', seasonal='add').fit().predict(start=start_time, end=future_time)
    y_hat_avg[y_hat_avg < 0] = 0
    y_hat_avg['forecast'] = y_hat_avg['forecast'].replace(
        np.nan, 0).round(0).astype(int)

    # Past data top 3
    top_dates = df_hour.sort_values(by=['counts'], ascending=False).head(3)
    vals = []
    for tgl, tot in zip(top_dates.index, top_dates["counts"]):
        val = "%d" % (tot)
        vals.append(val)
    top_dates['tgl'] = vals

    fig = go.Figure(data=go.Scatter(x=top_dates.index, y=top_dates['counts'],
                                    textposition='top center',
                                    textfont=dict(color='#233a77'),
                                    mode='markers+text',
                                    marker=dict(color='red', size=3),
                                    visible=True, name="Top 3 Total Failures",
                                    text=top_dates["tgl"]))

# fig.add_traces(data=go.Scatter(x=df_hour.index.astype(dtype=str),
#                                 y=df_hour['counts'],
#                                 visible = True,
#                                 marker_color='red',
#                                 name = "Failures (line)",
#                                 text="counts"))

    fig.add_traces(data=go.Bar(x=df_hour.index.astype(dtype=str),
                               y=df_hour['counts'], visible=True, name="Actual Total Failures",
                               marker_color='tomato', text=df_hour["counts"]))

# fig.add_traces(data=go.Scatter(x=y_hat_avg.index.astype(dtype=str),
#                           y=y_hat_avg['forecast'],visible = True, name = "Failure Forecast (line)",
#                           marker_color='orange', text="forecast"))

    fig.add_traces(data=go.Bar(x=y_hat_avg.index.astype(dtype=str),
                               y=y_hat_avg['forecast'], visible=True, name="Forecast",
                               marker_color='orange', text=y_hat_avg["forecast"]))

    fig.update_layout(bargap=0.01, showlegend=True, autosize=True, height=400, width=1200, title={"text": "Weekly Failures Trend per hour", "y": 0.97, "x": 0.5,
                                                                                                  "xanchor": "center", "yanchor": "top"})
    fig.update_xaxes(
        rangeslider_visible=False)
    fig.show()
    img = fig.to_image(format="png", engine="kaleido")

    summary = '<img alt="execution summary" id="box" style="height: 400px;width: 1200px !important;" src="data:image/png;base64, {}" >'.format(
        b64encode(img).decode("utf-8"))
    return summary


print(forecast_failures(df))



import requests
import json
import pandas
import os


def get_meta_data():
    url = "https://"+os.environ["cloudName"] + \
        ".app.perfectomobile.com/test-execution-management-webapp/rest/v1/metadata"

    payload = {}
    headers = {
        'Perfecto-Authorization': os.environ['securityToken'],
        'perfecto-tenantid': os.environ["cloudName"] + '-perfectomobile-com'
    }
    response = requests.request("GET", url, headers=headers, data=payload, timeout=30)
    result = json.loads(response.content)
    return result


def get_failure_reason_name_category(result):
    try:
        resultList = result["failureReasons"]
    except TypeError:
        print(result)
    if (len(resultList) > 0):
        return pandas.DataFrame(resultList, columns=['name', 'category'])
    else:
        return pandas.DataFrame()


async def get_failure_reason_category_table(df, total):
    # get category df
    result = get_meta_data()
    category_df = get_failure_reason_name_category(result)
    category_df.rename(
        columns={'name': 'Custom Failure Reasons', 'category': 'Category'}, inplace=True)
    df = category_df.merge(df, how="inner", on="Custom Failure Reasons").pivot_table(
        index=['Status', 'Category', 'Custom Failure Reasons'], aggfunc='size')
    df = pandas.DataFrame(df).sort_values(by=0, ascending=False).sort_values(
        by="Status", ascending=False).head(10).rename(columns={0: "Total"})
    df['✓%↑'] = round(df["Total"].div(
        total).mul(100).astype(float), 1).astype(str) + '%'
    failurecategorytable = {}


    failurecategorytable = df.to_html(
        table_id="category",
        index=True,
        render_links=True,
        escape=False,
    )
    return failurecategorytable




def prepare_report(client_logo, mobile_web, web, top_count, mobile_web_recommender, web_recommender, json):
    # https://codepen.io/dp_lewis/pen/WNZQzN
    html = str('''
    <!DOCTYPE html>
    <script src="http://code.jquery.com/jquery-1.11.0.min.js"></script>
    <META http-equiv="Content-Type" content="text/html; charset=UTF-8"> 
    <html>
    <head>
        <style>
        
            #combined {
                display: inline-flex;
                justify-content: center;
                flex-direction: row;
                overflow: scroll;
                justify-content: flex-start;
            }
            
            input[type="radio"] {
                visibility: hidden;
            }
            
            #1,#2,#3 {
                height:1px;
            }
            
            #label {
                background: #444;
                color: #fff;
                transition: transform 400ms ease-out;
                display: inline-grid;
                width: 100vw;
                height: 430px;
                position: relative;
                z-index: 1;
                text-align: center;
                line-height: 4vh;
                margin-top: -5px;
            }

            form {
                position: absolute;
                left: 0;
                right: 0;
                white-space: nowrap;
            }

            input {
                position: absolute;
            }

            input:nth-of-type(1):checked~#label:nth-of-type(1),
            input:nth-of-type(2):checked~#label:nth-of-type(2),
            input:nth-of-type(3):checked~#label:nth-of-type(3) {
                z-index: 0;
            }

            input:nth-of-type(1):checked~#label {
                transform: translate3d(0, 0, 0);
            }

            input:nth-of-type(2):checked~#label {
                transform: translate3d(-100%, 0, 0);
            }

            input:nth-of-type(3):checked~#label {
                transform: translate3d(-200%, 0, 0);
            }

            #label {
                background: rgba(232, 255, 227, 0.58);
                background-size: cover;
                font-size: 1rem;
            }

            #label:before,
            #label:after {
                color: white;
                display: block;
                background: rgba(22, 34, 73, 0.27);
                position: absolute;
                padding: 1rem;
                font-size: 2rem;
                height: 9rem;
                vertical-align: middle;
                line-height: 9rem;
                top: 52%;
                transform: translate3d(0, -50%, 0);
                cursor: pointer;
            }

            #label:before {
                content: ">";
                right: 100%;
                border-top-left-radius: 50%;
                border-bottom-left-radius: 50%;
            }

            #label:after {
                content: "<";
                left: 100%;
                border-top-right-radius: 50%;
                border-bottom-right-radius: 50%;
            }
        
            html {
                display:inline-grid;
            }
            
            body {
                font-family: sans-serif;
                font-size: 11px;
                margin: 0px;
                background: #f0f7f7bf;
                overflow: hidden;
            }

            table, th, td {
                border: 0px solid black;
                border-collapse: collapse;
                font-size: 11px;
            }


            table.center {
                margin-left: auto;
                margin-right: auto;
            }

            th,
            td {
                padding: 2px;
                font-weight: 100;
            }

            th {
                border-radius: 5px 5px 0px 0px;
                text-align: left;
                box-sizing: border-box;
                border: 2px solid #366e9936;
                font-size: 12px;
                background: #5e90d8;
                font-weight: 600;
                color: white;
                line-height: 150%;
                vertical-align: middle;
            }

            td {
                box-sizing: border-box;
                border: 2px solid #366e9936;
                position: relative;
                color: #200302;
                background: #ffffff;
                vertical-align: middle;
                font-size: 11px;
                line-height: 140%;
                text-align: left;
                border-radius: 10px 0px 10px 0px;
            }

            img {
                display: inline-flex;
                height: 100%;
            }

            #summary {
                text-align: center;
                background: #f0ffff;
            }

            #piechart,
            #browserChart {
                background-color: rgba(238, 240, 223, 0.68);
                display: block;
                width: 300px;
                height: 240px;
                margin: 5px;
                box-shadow: 0 0 60px rgb(80 130 50 / 33%);
            }
        
            #images {
                background: linear-gradient(to right, #ffb35a70, #E3F3FE, #E3F3FE, #aacfe8 90.33%, #ffb35a70);
                box-shadow: 0 1px 6px rgba(0, 0, 0, 0.12), 0 1px 4px rgba(0, 0, 0, 0.24);
                height: 35px;
                margin: auto;
                display: flex;
                justify-content: center;
                padding: 2px 1px 2px 10px;
            }

            #clientlogo {
                width: 100px;
                padding-left: 20px;
            }

            #perfecto {
                width: 100px;
            }

            #table {
                box-shadow: 0 10px 10px rgb(80 130 50 / 33%);
                overflow-y: scroll;
                word-wrap: break-word;
                margin-top: 0px;
                max-height: 533px;
                border: 2px solid #90cfe27d;
                border-radius: 6px;
                width: 100%;
            }

            #slide {
                display: inline-block;
                flex-direction: row;
                overflow: scroll;
                align-items: self-start;
                justify-content: flex-start;
                margin-top: 8px;
            }

            a {
                color: #04192d;
            }

            #heading {
                font-size: small;
                display: flex;
                flex-direction: row;
                align-items: center;
                justify-content: center;
                font-weight: bold;
                width: 100%;
                padding: 5px 0;
                text-align: center;
                cursor: pointer;
                color: white;
                background: #002e5dd1;
                height: 15px;
                border-radius: 5px 5px 0px 0px;
            }

            svg {
                overflow: scroll;
            }

            #box {
                box-shadow: 0 -4px 18px 0 rgba(60, 118, 191, 0.55);
                margin-right: 2px;
                margin-left: 2px;
                margin-top: 3px;
                margin-bottom: -2px;
                border: 1px solid #89def77d;
                width: 100%;
                object-fit: cover;
                border-radius: 6px;
            }

            #rect {
                border: 5px solid #12303947;
                border-radius: 10px;
                margin: 3px;
                display: inline-block;
                vertical-align:middle;
                background: #add8e6b8;
            }
            #jsoneditor {
                margin-top: 430px !important;
                width: 800px;
                height: 760px;  
                overflow: auto;
            }
            .jse-hidden-input-label{
              height: 0px  !important;
            }
        </style>
        </head>
        <title>PerfectReports - Analytics</title>

        <body>
            <div id="images">
                <img src='https://www.perfecto.io/sites/default/themes/custom/perfecto/logo.svg?height=57&width=200'
                    alt="Perfecto" id="perfecto" class="center">
                <img src="''' + client_logo + '''" alt="logo" id="clientlogo" class="center">
                <img src="https://fibbl.com/wp-content/uploads/2023/10/6322e2479751c370b289db18_Logo_Google_Analytics.svg.png" alt="google_logo" id="clientlogo" class="center">
            </div>
            
            <div id="summary">
            <form>
                <input type="radio" name="fancy" autofocus value="1" id="1" />
                <input type="radio" name="fancy" value="2" id="2" />
                <input type="radio" name="fancy" value="3" id="3" />
                <label id="label" for="1">
                    <div id="slide">
                        <div id="box">
                            <div id="heading">Web - Top ''' + top_count + '''</div>
                            <div id="combined">
                                <div id="rect">''' + web[0] + '''</div>
                                <div id="rect">''' + web[1] + '''</div>
                                <div id="rect">''' + web[2] + '''</div>
                               <div id="rect"><div id="heading">PS Recommendation</div>''' + web_recommender[0] + '''</div>
                            </div>
                        </div>
                    </div>
                </label>
                <label id="label" for="2">
                    <div id="slide">
                        <div id="box">
                            <div id="heading">iOS - Top ''' + str(top_count) + '''</div>
                            <div id="combined">
                                <div id="rect">''' + mobile_web[0] + '''</div>
                                <div id="rect">''' + mobile_web[1] + '''</div>
                                <div id="rect">''' + mobile_web[2] + '''</div>
                                <div id="rect"><div id="heading">PS Recommendation</div>''' + mobile_web_recommender[0] + '''</div>
                            </div>
                        </div>
                    </div>
                </label>
                <label id="label" for="3">
                    <div id="slide">
                        <div id="box">
                            <div id="heading">Android - Top ''' + top_count + '''</div>
                            <div id="combined">
                                <div id="rect">''' + mobile_web[3] + '''</div>
                                <div id="rect">''' + mobile_web[4] + '''</div>
                                <div id="rect">''' + mobile_web[5] + '''</div>
                                <div id="rect"><div id="heading">PS Recommendation</div>''' + mobile_web_recommender[1] + '''</div>
                            </div>
                        </div>
                    </div>
                </label>
            </form>
            </div>
            <div style="display: flex;justify-content: center;background: lightcyan;">
                <div id="jsoneditor"></div>
            </div>
            <script type="module">
                import { JSONEditor } from 'https://cdn.jsdelivr.net/npm/vanilla-jsoneditor/standalone.js'

                let content = {
                    text: undefined,
                    json: ''' + str(json) + '''
                    }

                    const editor = new JSONEditor({
                        target: document.getElementById('jsoneditor'),
                        props: {
                        content,
                        onChange: (updatedContent, previousContent, { contentErrors, patchResult }) => {
                            // content is an object { json: unknown } | { text: string }
                            console.log('onChange', { updatedContent, previousContent, contentErrors, patchResult })
                            content = updatedContent
                        }
                        }
                    })

                    // use methods get, set, update, and onChange to get data in or out of the editor.
                    // Use updateProps to update properties.
                    </script>
        </body>
        </html>
            ''')
    return html


import base64
import json
import sys
import numpy as np
import matplotlib.pyplot as plt
import datetime
import io
import base64
import os
from time import time
import pandas
import requests
import plotly.graph_objects as go
from base64 import b64encode


def hyperlink(df, url, from_col, to_col, first_df, first_item=False):
    videoURLS = []
    for ind in df.index:
        if first_item:
            videoURLS.append(
                first_df.loc[first_df[to_col] ==
                             df[to_col][ind], from_col].iloc[0]
            )
        else:
            videoURLS.append(df[from_col][ind])
    df[url] = videoURLS
    df[url] = df[url].apply(
        lambda x: "{0}".format(x)
    )
    for ind in df.index:
        df.loc[df[to_col].index == ind, to_col] = (
            '<a target="_blank" href="'
            + df[url][ind]
            + '">'
            + df[to_col][ind]
            + "</a>"
        )
    df = df.drop(url, axis=1)
    return df


async def topFailedTC(total, failed_df):
    failed_df = failed_df[~failed_df['name'].str.contains(
        "Interactive session")]
    # top failed TCs
    topfailedTCNames = (
        failed_df.groupby(["name"])
        .size()
        .reset_index(name="#Failed")
        .sort_values("#Failed", ascending=False)
        .head(10)
    )
    topfailedTCNames['%'] = round(topfailedTCNames["#Failed"].div(
        total).mul(100).astype(float), 1).astype(str) + '%'
    topfailedTCNames = hyperlink(
        topfailedTCNames, "Result", "reportURL", "name", failed_df, first_item=True)
    topfailedTCNames.columns = ["Failed Tests", "Total", "✓%↑"]
    topfailedtable = {}
    topfailedtable = topfailedTCNames.to_html(
        table_id="topfailedtests",
        index=False,
        render_links=True,
        escape=False,
    )
    return topfailedtable


async def topFailedReasons(total, failed_df):
    # top failed reasons
    topfailedTCNames = (
        failed_df.groupby(["Failure Message"])
        .size()
        .reset_index(name="#Failed")
        .sort_values("#Failed", ascending=False)
        .head(10)
    )
    topfailedTCNames['%'] = round(topfailedTCNames["#Failed"].div(
        total).mul(100).astype(float), 1).astype(str) + '%'
    topfailedTCNames = hyperlink(
        topfailedTCNames, "Result", "reportURL", "Failure Message", failed_df, first_item=True)
    topfailedTCNames.columns = ["Failure Messages", "Total", "✓%↑"]
    topfailedtable = {}
    topfailedtable = topfailedTCNames.to_html(
        table_id="topfailedmessages",
        index=False,
        render_links=True,
        escape=False,
    )
    return topfailedtable


def get_report_details(item, temp, name, criteria):
    if name + "=" in item:
        temp = str(item).split("=", 1)[1]
    return str(temp), criteria


def get_key_details(item, temp, name):
    if name + "=" in item:
        temp = str(item).split("=", 1)[1]
    return str(temp)


def fig_to_base64(fig):
    img = io.BytesIO()
    fig.savefig(img, format='png',
                bbox_inches='tight')
    img.seek(0)

    return base64.b64encode(img.getvalue())


async def create_summary_pie(df, title, column, value_column):
    status_df = df[column].value_counts().sort_index().reindex(
        ["BLOCKED", "FAILED", "PASSED", "UNKNOWN"], fill_value=0).rename_axis(column).to_frame()
    layout = go.Layout(
        autosize=True,
        showlegend=False,
        title_text=title,
        title_x=0.5,
        title_y=0.97,
    )
    PASSED = '#33d633fa'
    FAILED = '#ff5f41d6'
    BLOCKED = '#ffa50091'
    UNKNOWN = '#F2EBEF'
    colors = [BLOCKED, FAILED, PASSED, UNKNOWN]
    fig = go.Figure(data=[go.Pie(labels=status_df.index.tolist(),
                    values=status_df[value_column].values.tolist(), pull=[0, 0.05, 0.1, 0],
                    hole=.3, texttemplate="%{label}: %{value} <br>(%{percent})", insidetextorientation='horizontal', sort=True)], layout=layout)
    fig.update_traces(marker=dict(
        colors=colors, line=dict(color='white', width=3)))
    fig.update_layout(font=dict(family="sans-serif",
                      size=22, color="black"))
    img = fig.to_image(format="png", engine="kaleido")

    summary = '<img alt="execution summary" id="box" style="height: 200px;width: 240px !important;" src="data:image/png;base64, {}"'.format(
        b64encode(img).decode("utf-8"))
    return summary


def create_pie(df, title, column, value_column, style="height: 240px;width: 300px;"):
    status_df = df[column].value_counts(
    ).sort_index().rename_axis(column).to_frame()
    layout = go.Layout(
        autosize=True,
        showlegend=False,
        title_text=title,
        title_x=0.5,
        title_y=0.97,
    )
    fig = go.Figure(data=[go.Pie(labels=status_df.index.tolist(),
                    values=status_df[value_column].values.tolist(), pull=[0.1, 0.05, 0.02, 0],
                    hole=.3, texttemplate="%{label}: %{value} <br>(%{percent})", insidetextorientation='horizontal', sort=True)], layout=layout)
    fig.update_traces(marker=dict(line=dict(color='white', width=3)))
    fig.update_layout(font=dict(family="sans-serif",
                      size=22, color="black"))
    img = fig.to_image(format="png", engine="kaleido")

    summary = '<img alt="execution summary" id="box" style="'+style+'" src="data:image/png;base64, {}" >'.format(
        b64encode(img).decode("utf-8"))
    return summary


def create_top_pie(df, title, column, value_column, top, style="height: 240px;width: 300px;"):
    status_df = df[column].value_counts().reset_index(
        name=value_column).sort_values(value_column, ascending=False).head(top)
    layout = go.Layout(
        autosize=True,
        showlegend=False,
        title_text=title,
        title_x=0.5,
        title_y=0.97,
    )
    labels = []
    if column not in status_df.columns:
        labels = status_df['index'].tolist()
    else:
        labels = status_df[column].tolist()
    fig = go.Figure(data=[go.Pie(labels=labels,
                    values=status_df[value_column].values.tolist(), pull=[0.1, 0.05, 0.02, 0],
                    hole=.3, texttemplate="%{label}: %{value} <br>(%{percent})", insidetextorientation='horizontal', sort=True)], layout=layout)
    fig.update_traces(marker=dict(line=dict(color='white', width=3)))
    fig.update_layout(font=dict(family="sans-serif",
                      size=22, color="black"))
    img = fig.to_image(format="png", engine="kaleido")

    summary = '<img alt="execution summary" id="box" style="'+style+'" src="data:image/png;base64, {}" >'.format(
        b64encode(img).decode("utf-8"))
    return summary


"""
    Dictionary
"""


class my_dictionary(dict):
    def __init__(self):
        self = dict()

    def add(self, key, value):
        self[key] = value


"""
    flattens the json
"""


def flatten_json(nested_json, exclude=[""]):
    """Flatten json object with nested keys into a single level.
        Args:
            nested_json: A nested json object.
            exclude: Keys to exclude from output.
        Returns:
            The flattened json object if successful, None otherwise.
    """
    out = {}

    def flatten(x, name="", exclude=exclude):
        if type(x) is dict:
            for a in x:
                if a not in exclude:
                    flatten(x[a], name + a + "/")
        elif type(x) is list:
            i = 0
            for a in x:
                flatten(a, name + str(i) + "/")
                i += 1
        else:
            out[name[:-1]] = x
    flatten(nested_json)
    return out


def df_formatter(df, jobName, jobNumber):
    if len(df) < 1:
        raise Exception("Unable to find any matching executions!")
    try:
        df["startTime"] = pandas.to_datetime(df["startTime"], unit="ms")
        df["startTime"] = (
            df["startTime"].dt.tz_localize(
                "utc").dt.tz_convert(None)
        )
        df["startTime"] = df["startTime"].dt.strftime("%Y-%m-%d %H:%M:%S")
    except:
        pass
    try:
        df.loc[df["endTime"] < 1, "endTime"] = int(round(time() * 1000))
        df["endTime"] = pandas.to_datetime(df["endTime"], unit="ms")
        df["endTime"] = (
            df["endTime"].dt.tz_localize(
                "utc").dt.tz_convert(None)
        )
        df["endTime"] = df["endTime"].dt.strftime("%Y-%m-%d %H:%M:%S")
    except:
        pass
    if "month" not in df.columns:
        df["month"] = pandas.to_datetime(
            df["startTime"], format="%Y-%m-%d %H:%M:%S"
        ).dt.to_period("M")
    if "startDate" not in df.columns:
        df["startDate"] = pandas.to_datetime(
            pandas.to_datetime(df["startTime"], format="%Y-%m-%d %H:%M:%S")
            .dt.to_period("D")
            .astype(str)
        )
    if "week" not in df.columns:
        df["week"] = pandas.to_datetime(df["startDate"].dt.strftime("%Y/%m/%d")) - df[
            "startDate"
        ].dt.weekday.astype("timedelta64[D]")
    if "Duration" not in df.columns:
        df["Duration"] = pandas.to_datetime(
            df["endTime"], format='mixed') - pandas.to_datetime(df["startTime"], format='mixed')
        df["Duration"] = df["Duration"].dt.seconds
        df["Duration"] = pandas.to_datetime(df["Duration"], unit="s").dt.strftime(
            "%H:%M:%S"
        )
    if "failureReasonName" not in df.columns:
        df["failureReasonName"] = ""
    # Filter only job and job number if dates are parameterized as well but show full histogram
    if jobNumber != "" and jobName != "":
        if ";" in jobNumber:
            df = df[df["job/number"].isin(jobNumber.split(";"))]
        else:
            df = df[df["job/number"].astype(str) == jobNumber]
    return df


def payloadJobAll(reportTags, oldmilliSecs, current_time_millis, jobName, jobNumber, page, internal, boolean):
    payload = my_dictionary()
    if oldmilliSecs != 0:
        payload.add("startExecutionTime[0]", oldmilliSecs)
    if reportTags != "":
        for i, reportTaging in enumerate(reportTags.split(";")):
            if internal == "false":
                payload.add("tags[" + str(i) + "]", reportTaging)
            else:
                payload.add("tags", reportTaging)
    if current_time_millis != 0:
        payload.add("endExecutionTime[0]", current_time_millis)
    if internal == "false":
        payload.add("_page", page)
    if jobName != "":
        if jobName != "All Jobs":
            if jobName != "Perfecto Integrations":
                for i, job in enumerate(jobName.split(";")):
                    if internal == "false":
                        payload.add("jobName[" + str(i) + "]", job)
                    else:
                        payload.add("jobName", job)
    if jobNumber != "" and boolean:
        for i, jobNumber in enumerate(jobNumber.split(";")):
            if internal == "false":
                payload.add("jobNumber[" + str(i) + "]", jobNumber)
            else:
                payload.add("jobNumber", int(jobNumber))
    print(str(payload))
    return payload

def pastDateToMS(startDate, daysOlder):
    dt_obj = datetime.datetime.strptime(
        startDate + " 00:00:00,00", "%Y-%m-%d %H:%M:%S,%f"
    ) - datetime.timedelta(days=daysOlder)
    millisec = dt_obj.timestamp() * 1000
    oldmilliSecs = round(int(millisec))
    return oldmilliSecs


def retrieve_tests_executions(daysOlder, jobName, jobNumber, page, startDate, endDate, reportTag, skip, pageSize, internal):
    current_time_millis = 0
    oldmilliSecs = 0
    if endDate != "":
        if "-" not in endDate:
            current_time_millis = endDate
        else:
            endTime = datetime.datetime.strptime(
                str(endDate) + " 23:59:59,999", "%Y-%m-%d %H:%M:%S,%f"
            )
            millisec = endTime.timestamp() * 1000
            current_time_millis = round(int(millisec))
    if startDate != "":
        if "-" not in startDate:
            oldmilliSecs = startDate
        else:
            oldmilliSecs = pastDateToMS(startDate, daysOlder)
    if jobNumber != "" and jobName != "" and startDate != "" and endDate != "":
        fields = payloadJobAll(
            reportTag, oldmilliSecs, current_time_millis, jobName, jobNumber, page, internal, False
        )
    else:
        fields = payloadJobAll(
            reportTag, oldmilliSecs, current_time_millis, jobName, jobNumber, page, internal, True
        )

    if internal == "true":
        payload = {
            "filter": {
                "fieldNameToSearchFilter": {},
                "fields": fields,
                "excludedFields": {}
            },
            "layout": [
                "job",
                "id",
                "reportUrl",
                "version",
                "name",
                "status",
                "startTime",
                "failureReason",
                "endTime",
                "platforms",
                "tags",
                "failureReason",
                "automationFramework",
                "cleanException",
                "message",
                "errorAnalysis"
            ],
            "skip": skip,
            "pageSize": pageSize
        }
        url = "https://" + os.environ["cloudName"] + \
            ".app.perfectomobile.com"
        api_url = url + "/test-execution-management-webapp/rest/v1/test-execution-management/search"

        # creates http get request with the url, given parameters (payload) and header (for authentication)
        r = requests.request("POST",
                             api_url, data=json.dumps(payload), headers={
                                 "PERFECTO_AUTHORIZATION": os.environ["securityToken"],
                                 "content-type": "application/json",
                                 "perfecto-tenantId": os.environ["cloudName"] + "-perfectomobile-com"})
    else:
        url = "https://" + os.environ["cloudName"] + \
            ".reporting.perfectomobile.com"
        api_url = url + "/export/api/v1/test-executions"

        # creates http get request with the url, given parameters (payload) and header (for authentication)
        r = requests.get(
            api_url, params=fields, headers={
                "PERFECTO_AUTHORIZATION": os.environ["securityToken"],
                "Content-Type": "application/json",
                "Perfecto-TenantId": os.environ["cloudName"] + "-perfectomobile-com"}
        )
    print(str(r.url))  # debug
    return r.content

def prepare_device_coverage(df):
    if 'model' not in df.columns:
        df['model'] = np.NaN
    if 'browserType' not in df.columns:
        df['browserType'] = np.NaN
    if 'browserVersion' not in df.columns:
        df['browserVersion'] = np.NaN
        # append version to device/browser version columns and combine them
    df.loc[df['browserType'].notna(), 'browserType'] = df.loc[df['browserType'].notna(
    ), 'browserType'].astype(str)+'-v'
    df['browser'] = df["browserType"] + df["browserVersion"].astype(str)
    df.loc[df['model'].notna(), 'model'] = df.loc[df['model'].notna(),
                                                  'model'].astype(str)+'-v'
    df['mobile'] = df['model'].astype(
        str).replace("-v", '') + df['Version'].astype(str)
    df['osCoverage'] = df[['browser', 'mobile']
                          ].stack().groupby(level=0).agg(''.join)
    df['osCoverage'] = df['osCoverage'].str.replace(
        r'nan.*', '', regex=True).astype('str')
    df = df.drop(df[df['osCoverage'] == 'nan'].index)
    if (len(df['osCoverage'].value_counts()) > 0):
        version_items = create_top_pie(
            df, "", "osCoverage", "Total", 5, "height: 200px;width: 290px;")
        version_items = '<div id="box"><div id="heading">Top 5 Browsers / Devices</div><div>' + \
            version_items + ' </div></div>'
    else:
        version_items = ""
    return df, version_items


def get_resources(truncated, reportTag, page, daysOlder, resources, jobName, jobNumber, startDate, endDate, internal):
    skip, pageSize = 0, 200
    # executionList = {'executionList': '1'} # workaround for turncated issue
    # while len(executionList) > 0: # workaround for turncated issue
    while truncated == True:
        executions = retrieve_tests_executions(
            daysOlder, jobName, jobNumber, page, startDate, endDate, reportTag, skip, pageSize, internal)
        # Loads JSON string into JSON object
        executions = json.loads(executions)
        if "{'userMessage': 'Failed decoding the offline token:" in str(executions):
            raise Exception(
                "please change the offline token for your cloud")
        if "userMessage': 'Missing Perfecto-TenantId header" in str(executions):
            raise Exception("Check the cloud name and security tokens")
        if "userMessage': 'Time period is not in supported range" in str(executions):
            raise Exception(
                "Time period is not in supported range. Check your startDate parameter")
        try:
            if internal == "true":
                executionList = executions["items"]
            else:
                executionList = executions["resources"]
        except TypeError:
            print(executions)
            raise Exception(
                "Unable to find matching records for: "
                + str(sys.argv[2], sys.argv[3])
                + ", error:"
                + str(executions["userMessage"])
            )
            sys.exit(-1)
        if len(executionList) == 0:
            return resources
        else:
            if internal == "true":
                resources.extend(executionList)
                skip += pageSize
            else:
                metadata = executions["metadata"]
                truncated = metadata["truncated"]
                if page >= 1:
                    resources.extend(executionList)
                else:
                    resources.append(executionList)
            print('page: ' + str(page) + ', page count: ' +str(len(executionList)))
            page += 1
    return resources

async def prepare_custom_failure_graph(failed_df):
    if (len(failed_df) > 0):
        return create_top_pie(
            failed_df, "", "Custom Failure Reasons", "count", 5, "height: 200px;width: 300px;")
    else:
        return ""


async def prepare_module_graph(df):
    # https://towardsdatascience.com/100-stacked-charts-in-python-6ca3e1962d2b
    df = df[~df['Module'].isin(['-'])]
    cross_tab = pandas.crosstab(index=df['Module'],
                                columns=df['Status'])
    cross_tab_prop = pandas.crosstab(index=df['Module'],
                                     columns=df['Status'],
                                     normalize="index")
    ax = cross_tab_prop.sort_values('Module', ascending=False).plot(kind='barh',
                                                                    stacked=True,
                                                                    color={'PASSED': '#33d633fa',
                                                                           'FAILED': '#ff5f41b5',
                                                                           'BLOCKED': '#ffa50091',
                                                                           'UNKNOWN': '#F2EBEF'},
                                                                    )  # figure size figsize=(4, 2.5)
    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.axes.get_xaxis().set_visible(False)
    ax.xaxis.set_tick_params(labelsize=10)
    ax.yaxis.set_tick_params(
        labelsize=9, labelcolor="black", direction='out', which='major')
    ax.get_legend().remove()
    # plt.rcParams["font.weight"] = "bold"
    plt.xlabel("")
    plt.ylabel("")
    plt.tight_layout()
    F = plt.gcf()
    Size = F.get_size_inches()
    F.set_size_inches(Size[0]*.95, Size[1]*.7, forward=True)
    for n, x in enumerate([*cross_tab.sort_values('Module', ascending=False).index.values]):
        for (proportion, count, y_loc) in zip(cross_tab_prop.loc[x],
                                              cross_tab.loc[x],
                                              cross_tab_prop.loc[x].cumsum()):
            # time.sleep(.7)
            plt.text(x=(y_loc - proportion) + (proportion / 2),
                     y=n - 0.11,
                     s=f'{count}\n({np.round(proportion * 100, 1)}%)',
                     color="#36220e",
                     fontsize=9)
    fig = 'tags.png'
    plt.savefig(fig, bbox_inches='tight')
    with open(fig, "rb") as fig:
        encoded = base64.b64encode(fig.read())
        tags_base64 = '<img src="data:image/png;base64, {}">'.format(
            encoded.decode('utf-8'))

    return tags_base64


async def prepare_tags_table(total, failed_df):
    tags_df = failed_df.pivot_table(
        index=['Owner', 'Module', 'Failed Steps'], aggfunc='size')
    tags_df = pandas.DataFrame(tags_df)
    tags_df = tags_df.sort_values([0]).tail(10)
    tags_df = tags_df.rename(
        columns={
            0: "Count",
        })
    tags_df.sort_values(['Count'],
                        ascending=[False], inplace=True)
    tags_df['%'] = round(tags_df["Count"].div(
        total).mul(100).astype(float), 1).astype(str) + '%'
    tags_df_table = tags_df.to_html(
        table_id="tags",
        render_links=True,
        escape=False,
    )

    return tags_df_table


async def prepare_failed_table(failed_df, internal):
    failed_df["Video"] = '<img id="vidlink"/>'
    if internal == "false":
        failed_df = hyperlink(failed_df, "vidResult",
                              "videos/0/downloadUrl", "Video", "", False)
        failed_df = failed_df[["Owner", "Module", "name", "Custom Failure Reasons",
                               "Failure Message", "Failed Steps", "Video"]]
    else:
        failed_df = failed_df[["Owner", "Module", "name", "Custom Failure Reasons",
                               "Failure Message", "Failed Steps"]]
    failed_df = failed_df.rename(
        columns={
            "name": "Failed Test"
        }
    )
    failedTable = failed_df.to_html(
        table_id="table",
        index=False,
        render_links=False,
        escape=False,
    )

    return failedTable


def format_df(df, jobName, jobNumber, automation_owners, internal):
    df = df_formatter(df, jobName, jobNumber)
    df["platforms/0/deviceType"] = df["platforms/0/deviceType"].fillna(
        "Others")
    df["platforms/0/os"] = df["platforms/0/os"].fillna("Others")
    if "failureReasonName" not in df.columns:
        df["failureReasonName"] = ""
    if internal == "true":
        df = df.rename(
            columns={"failureReason/name": "Custom Failure Reasons"})
    else:
        df = df.rename(
            columns={"failureReasonName": "Custom Failure Reasons"})
    df = df.rename(
        columns={
            "platforms/0/deviceType": "Platform",
            "platforms/0/os": "OS",
            "status": "Status",
            "errorAnalysis/cleanException": "cleanException",
            "platforms/0/osVersion": "Version",
            "tags/0": "Module",
            "platforms/0/browserInfo/browserVersion": "browserVersion",
            "platforms/0/browserInfo/browserType": "browserType",
            "platforms/0/mobileInfo/model": "model"
        }
    )
    del_cols = ["externalId", 'platforms/0/mobileInfo/imei',
                'platforms/0/mobileInfo/distributor', 'platforms/0/mobileInfo/firmware', 'triggerType',
                'failureReason/id', 'failureReason/isSystemAssigned', 'errorAnalysis/normalizedCleanException']
    df.drop(del_cols, axis=1, errors='ignore', inplace=True)
    df.reset_index(drop=True, inplace=True)
    # update owners
    if 'Module' not in df.columns:
        df['Module'] = '-'
    df["Owner"] = "-"
    df['Module'] = df['Module'].fillna("-")
    try:
        data = json.loads(str(automation_owners))
        for i in data:
            for j in data[i]:
                df.loc[df['Module'] == j, "Owner"] = i
    except Exception as e:
        pass

    # sort properly
    df.sort_values(['Owner', 'Module', 'name'],
                   ascending=[True, True, True], inplace=True)
    df.reset_index(drop=True, inplace=True)
    return df


def generateExcel(df):
    df.to_csv("output.csv", index=False)


def get_os_versions_list(devices):
    osVersions = []
    for device in devices['handsets']['handset']:
        osVersions.append(str(device['osVersion']))
    return list(set(osVersions))


def find_nearest_decimal(decimal, array_list):
    whole_number = int(decimal)
    nearest_decimal = None
    min_difference = float('inf')

    for number in array_list:
        integer_part = int(float(number))
        difference = abs(float(integer_part) - float(whole_number))
        if difference < min_difference:
            min_difference = difference
            nearest_decimal = number
    return nearest_decimal


def get_os_version_major(osVersions):
    return ['.'.join(version.split(".")[:2]) for version in osVersions]
